# This file is part of CloudControl.
#
# CloudControl is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CloudControl is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with CloudControl.  If not, see <http://www.gnu.org/licenses/>.


""" In-database tags representation.
"""

from abc import ABCMeta, abstractproperty
from datetime import datetime, timedelta


class BaseTagInterface(object):

    """ Base class for all tag interfaces.
    """

    __metaclass__ = ABCMeta


class BaseTag(object):

    """ Base class for a tag in the database.

    :param name: name of the tag
    """

    interface = None  # Define the tag fetching interface of the tag

    def __init__(self, name, **kwargs):
        self._name = name

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.name)

    def __hash__(self):
        return hash(self.name)

    #
    # Publics:
    #

    @property
    def name(self):
        return self._name


#
# Static tags
#

class StaticTagInterface(BaseTagInterface):

    """ An interface used to get static value stored by tags.
    """

    @abstractproperty
    def value(self):
        """ The value of the tag.
        """

    @value.setter
    def value(self, value):
        """ Set the value of the tag.
        """


class StaticTag(BaseTag):

    """ Static tag in the database.
    """

    interface = 'static'

    def __init__(self, name, value='', **kwargs):
        super(StaticTag, self).__init__(name, **kwargs)
        self._value = value

    def __repr__(self):
        return '<%s %s (value=%r)>' % (self.__class__.__name__, self.name, self.value)

    #
    # Implentation of StaticTagInterface:
    #

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value


StaticTagInterface.register(StaticTag)

class CallbackTag(BaseTag, StaticTagInterface):

    """ A tag which is defined after a call to a callback.
    """

    interface = 'static'

    def __init__(self, name, callback, ttl=None, extra={}, **kwargs):
        super(CallbackTag, self).__init__(name, **kwargs)
        self._callback = callback
        self._callback_extra = extra
        self._ttl = ttl
        self._cache_value = ''
        self._cache_last_update = None

    def __repr__(self):
        if self._cache_last_update is not None and self._ttl is not None:
            dt = self.ttl - (datetime.now() - self._cache_last_update)
            expire = (dt.microseconds + (dt.seconds + dt.days * 24 * 3600) * 10**6) / 10.0**6
        else:
            expire = None
        return '<%s %s (ttl=%r expire=%r cached=%r cb=%r)>' % (self.__class__.__name__,
                                                               self.name,
                                                               self._ttl,
                                                               expire,
                                                               self._cache_value,
                                                               self._callback.__name__)

    #
    # Implentation of StaticTagInterface:
    #

    @property
    def value(self):
        # Update the cached value if it has expired or not retrieved yet:
        if self.ttl is not None or (self.ttl is None and self._cache_last_update is None):
            if self._cache_last_update is None or datetime.now() - self._cache_last_update > self.ttl:
                self.value = str(self.callback(**self._callback_extra))
        return self._cache_value

    @value.setter
    def value(self, value):
        self._cache_last_update = datetime.now()
        self._cache_value = value

    #
    # Publics:
    #

    @property
    def callback(self):
        return self._callback

    @callback.setter
    def callback(self, callback):
        self._callback = callback

    @property
    def ttl(self):
        return timedelta(seconds=self._ttl) if self._ttl is not None else None

    @ttl.setter
    def ttl(self, ttl):
        self._ttl = ttl

    def invalidate(self):
        self._cache_last_update = None
